
/* simple-file.cc */

// #include "clang/Basic/SourceManager.h"
// #include "clang/Driver/Compilation.h"
#include "clang/CodeGen/CodeGenAction.h"
#include "clang/Driver/Driver.h"
#include "clang/Frontend/CompilerInvocation.h"
#include "clang/Frontend/CompilerInstance.h"
#if defined(LLVM_2_8) || defined(LLVM_3_0) || defined(LLVM_3_1)
   #include "clang/Frontend/DiagnosticOptions.h"
#else
   #include "clang/Basic/DiagnosticOptions.h"
#endif
// #include "clang/Frontend/FrontendActions.h"
// #include "clang/Frontend/FrontendDiagnostic.h"
#include "clang/Serialization/ASTReader.h"
#include "clang/Frontend/TextDiagnosticPrinter.h"

#if defined(LLVM_2_8) || defined(LLVM_3_0) || defined(LLVM_3_1) || defined(LLVM_3_2)
   #include "llvm/LLVMContext.h"
   #include "llvm/Module.h"
#else
   #include "llvm/IR/Module.h"
#endif

// #include "llvm/ADT/OwningPtr.h"
#include "llvm/ExecutionEngine/ExecutionEngine.h"

#include "llvm/Support/MemoryBuffer.h"
#include "llvm/Support/ManagedStatic.h"
#ifdef LLVM_2_8
   #include "llvm/System/Host.h"
   #include "llvm/Target/TargetSelect.h"
#else
   #include "llvm/Support/Host.h"
   #include "llvm/Support/TargetSelect.h"
#endif

#include "clang/Frontend/Utils.h"

#include "llvm/ExecutionEngine/JIT.h"
// important, otherwise JIT is not linked
// see  http://stackoverflow.com/questions/3440130/llvm-code-generation-causing-seg-fault

// #include <stdio.h>
#include <iostream>
#include <stdexcept> // class std::runtime_error

using namespace std;

/* ---------------------------------------------------------------------- */

void info (string msg)
{
   cout << msg << endl;
}

void error (string msg)
{
   cerr << msg << endl;
   // throw new std::runtime_error (msg);
}

/* ---------------------------------------------------------------------- */

void execute (const char * source)
{
   static const char* argv [] = { "program", "-x", "c++",
                                  // "-fsyntax-only",
                                  // "-include-pch", "_output/example4.pch",
                                  // "-include-pth", "output/tgbutton.pth",
                                  // "-I/opt/root-5.26.00/include",
                                  // "-Wl,-rpath", "-Wl,/opt/root-5.26.00/lib",
                                  // "-lGui",
                                  #if 1
                                  "-I/usr/include",
                                  // "-I/usr/lib/gcc/i586-manbo-linux-gnu/4.3.2/include",
                                  "-I/usr/lib/gcc/i686-redhat-linux/4.9.2/include",
                                  #endif
                                  "string-input",
                                  // "-remap-file", "string-input;input/sample.cc"
                                  };

   static const int argc = sizeof (argv) / sizeof (argv[0]);

   #if defined(LLVM_2_8) || defined(LLVM_3_0) || defined(LLVM_3_1)
      clang::TextDiagnosticPrinter * diagClient =
         new clang::TextDiagnosticPrinter (llvm::errs(), clang::DiagnosticOptions ());
   #else
      llvm::IntrusiveRefCntPtr<clang::DiagnosticOptions> diagOpts = new clang::DiagnosticOptions();
      clang::TextDiagnosticPrinter * diagClient =
         new clang::TextDiagnosticPrinter(llvm::errs(), &*diagOpts);
   #endif

   #ifdef LLVM_2_8
      clang::Diagnostic diags (diagClient);
   #else
      llvm::IntrusiveRefCntPtr<clang::DiagnosticIDs> diagID (new clang::DiagnosticIDs());
      #if defined(LLVM_3_0) || defined(LLVM_3_1)
         clang::DiagnosticsEngine diags(diagID, diagClient);
      #else
         clang::DiagnosticsEngine diags(diagID, &*diagOpts, diagClient);
      #endif
   #endif

   clang::driver::Driver driver ("simple",
                                 #if defined(LLVM_2_8) || defined(LLVM_3_0)
                                 llvm::sys::getHostTriple(),
                                 #elif defined(LLVM_3_1) || defined(LLVM_3_2)
                                 llvm::sys::getDefaultTargetTriple(),
                                 #else
                                 llvm::sys::getProcessTriple(),
                                 #endif
                                 "a.out",
                                 #if defined(LLVM_2_8) || defined(LLVM_3_0) || defined(LLVM_3_1) || defined(LLVM_3_2)
                                 /* IsProduction = */ false,
                                 #endif
                                 #ifdef LLVM_2_8
                                 /* CXXIsProduction = */ false,
                                 #endif
                                 diags);

   driver.setTitle ("clang interpreter");

   clang::CompilerInvocation * invocation = new clang::CompilerInvocation;
   clang::CompilerInvocation::CreateFromArgs
       (*invocation,
        argv + 1, // skip program name
        argv + argc,
        diags);

   // Create a compiler instance to handle the actual work.
   clang::CompilerInstance CI;
   #ifdef LLVM_2_8
      CI.setLLVMContext (new llvm::LLVMContext);
   #endif
   CI.setInvocation (invocation);

   // Create memmory buffer with source text
   llvm::MemoryBuffer * buffer = llvm::MemoryBuffer::getMemBufferCopy (source, "SIMPLE_BUFFER");
   if (buffer == NULL)
      error ("Failed to create memory buffer");

   // Remap auxiliary name "string-input" to memory buffer
   clang::PreprocessorOptions & opts = CI.getInvocation().getPreprocessorOpts();
   opts.addRemappedFile ("string-input", buffer);

   // Create the compilers actual diagnostics engine.
   #if defined(LLVM_2_8) || defined(LLVM_3_0) || defined(LLVM_3_1) || defined(LLVM_3_2)
      CI.createDiagnostics (argc, const_cast <char**> (argv));
   #else
      CI.createDiagnostics ();
   #endif
   if (! CI.hasDiagnostics ())
      error ("No diagnostics");

   // Create and execute the frontend to generate an LLVM bitcode module.
   clang::CodeGenAction * action = new clang::EmitLLVMOnlyAction();
   if (! CI.ExecuteAction (*action))
      error ("Cannot execute action");

   llvm::Module * module = action->takeModule ();
   if (module == NULL)
      error ("No module");

   /* execute */

   llvm::InitializeNativeTarget ();

   std::string Error;
   llvm::ExecutionEngine * engine =
      llvm::ExecutionEngine::createJIT (module, &Error);
   if (engine == NULL)
     error ("Unable to make execution engine: " + Error);

   llvm::Function * func = module->getFunction ("main");
   if (func == NULL)
     error ("Function not found in module");

   std::vector <std::string> params;
   params.push_back (module->getModuleIdentifier ());

   engine->runFunctionAsMain (func, params, /* envp = */ NULL);

   /* clean up */

   // NO delete module;
   delete engine;
   delete action;
   // NO delete invocation;
   llvm::llvm_shutdown ();
}

/* ---------------------------------------------------------------------- */

// see: clang/examples/clang-interpreter/main.cpp
// see: clang/tools/driver/cc1_main.cpp

int main (int argc, const char * *argv)
{
   execute ("#include <stdio.h> \n"
            "int main () \n"
            "{ \n"
            "int i = 1; \n"
            "i ++; \n"
            // "i = f3 (1, 2); \n"
            "printf (\"Hello from string, i = %d \\n\", i);"
            "return 0; \n"
            "} \n");
  return 0;
}
